#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>
#include "ASL_user.h"

#define FALSE 0
#define TRUE 1
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define ZERO ((double) 0)

double Timer ( void ) ;

typedef struct SVMproblem_struct /* passed to user's evaluation routines */
{
    float      *Ax ; /* numerical values in matrix */
    short      *Ai ; /* row indices */
    int        *Ap ; /* pointers to start of each column in Ax, Ai */
    short       *a ; /* label vector (+- 1) */
    float       *A ; /* dense A matrix */
    int        *AP ; /* pointers to start of each column in A */
    int          n ; /* problem dimension */
    double  lo_cut ; /* x_i < lo_cut => x_i at lower bound */
    double   *temp ; /* temporary storage when computing A times vector */
    int   nfeature ; /* number of features */
    short Ax_is_one ; /* TRUE = all elements of Ax are 1 */
    short   kernel ; /* 0 = linear, 1 = poly, 2 = rbf, 3 = sig */
    short   degree ; /* parameter in kernel */
    float    gamma ; /* parameter in kernel */
    int        mem ; /* location of free space in A */
    double val_time ; /* time spent evaluating objective function */
} SVMproblem ;

void read_problem
(
    const char *filename
) ;

float sparse_dot
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

float binary_sparse_dot
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

float sparse_norm2
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

float binary_sparse_norm2
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

void scalar_init
(
    double  *x,  /* numerical values */
    double   s,  /* scalar */
    int      n   /* length of vectors */
) ;

void int_init
(
    int *x,  /* numerical values */
    int  s,  /* scalar */
    int  n   /* length of vectors */
) ;

void dense_daxpy
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
) ;

void get_col
(
    int k   /* build k-th column of A */
) ;

void SVMlinearAd
(
    ASLobjective *data
) ;

void SVMnonlinearAd
(
    ASLobjective *data
) ;

SVMproblem PB ;

int main (int argc, char **argv)
{
    short kernel, *Ai ;
    int i, j, k, n, nfeature, Apn, nlo, nhi, nfree, ncol, *Ap ;
    long N ;
    float *A, *Ax ;
    double b, C, lo, hi, t, tic, tic0, tol, xj, atx, *x, *g0, *temp ;
    ASLstat Stat ;
    ASLparm Parm ;
    C  = atof (argv [3]) ;
    printf ("\n\nproblem: %s // kernel: %s // C: %e\n",
             argv [1]+9, argv [2], C) ;
    tic = Timer () ;
    tic0 = tic ;
    read_problem (argv [1]) ;
    printf ("time to read data: %e\n", Timer () - tic) ;
    n = PB.n ;
    g0 = malloc (n*sizeof (double)) ;
    x = (double *) malloc (n*sizeof (double)) ;
    /* convert problem to standard form with a_i = 1 for all i
       x_i = -y_i + C if a_i = -1
       starting guess:    x_i = 0 for all i or
                          y_i = C if a_i = -1
                          x_i = y_i = 0 otherwise
       starting gradient: g_i = +1 if a_i = -1
                          g_i = -1 otherwise */
    b = 0. ;
    for (j = 0; j < n; j++)
    {
        if ( PB.a [j] < 0 )
        {
            x [j] = C ;
            b += C ;
            g0 [j] = +1 ;
        }
        else
        {
            x [j] = 0 ;
            g0 [j] = -1 ;
        }
    }
    PB.val_time = 0. ;
    N = (long) n ;
    nfeature = PB.nfeature ;
    Ap = PB.Ap ;
    Ai = PB.Ai ;
    Ax = PB.Ax ;
    Apn = Ap [n] ;
    printf ("number of features: %i\n", PB.nfeature) ;
    printf ("nonzeros in data: %i\n", Apn) ;
    printf ("problem dimension: %i\n", n) ;
    fflush (stdout) ;
 
    printf ("matrix to be created in format %s\n", argv [2]) ;
    fflush (stdout) ;
    kernel = 0 ;
    if ( strcmp (argv [2],"linear") ) /* if not linear, compute A */
    {
        if      ( !strcmp (argv [2], "poly") ) kernel = 1 ;
        else if ( !strcmp (argv [2], "rbf") )  kernel = 2 ;
        else if ( !strcmp (argv [2], "sig") )  kernel = 3 ;
        else
        {
            printf ("%s not a valid kernel\n", argv [2]) ;
            exit (0) ;
        }
    }
    PB.kernel = kernel ;

    if ( !strcmp (argv [2],"linear") )
    {
        temp = PB.temp = (double *) malloc (nfeature*sizeof (double)) ;
    }
    else
    {
        A = PB.A = (float *) malloc (N*N*sizeof (float)) ;
        PB.AP = (int *) malloc (N*sizeof (int)) ; /*pointer to location of col*/
        int_init (PB.AP, -1, n) ;
        PB.degree = 3 ;
        PB.gamma = 1/ (float) nfeature ;
        PB.mem = 0 ;                              /* free memory */
    }

    /* setup for ASL */
    lo = 0. ;
    hi = C ;
    tol = 1.e-3 ;
/*  if ( C > 9 ) tol = 1.e-2 ;*/
    ASLdefault (&Parm) ;

/*  Parm.maxits = 700 ; */
    Parm.PrintLevel = 0 ;
    Parm.lo_cut = 1.e-12 ;
    Parm.hi_cut = 1.e-12 ;
    Parm.QuadCost = TRUE ;
    Parm.a_is_1 = TRUE ;
    Parm.QuadLazy = FALSE ;
    Parm.bnd_is_scalar = TRUE ;
    if ( strcmp (argv [2],"linear") )
    {
        Parm.Subspace = TRUE ;
        Parm.nsub = 50 ;
    }

    /* usually tol = 1.e-3 */
    tic = Timer () ;
    if ( !strcmp (argv [2],"linear") )
    {
        printf ("call linear solver\n") ;
        fflush (stdout) ;
        ASL (x, &lo, &hi, NULL, b, g0, n, tol, &Stat, &Parm,
             SVMlinearAd, NULL, NULL, NULL) ;
    }
    else
    {
        printf ("call nonlinear solver\n") ;
        fflush (stdout) ;
        ASL (x, &lo, &hi, NULL, b, g0, n, tol, &Stat, &Parm,
             SVMnonlinearAd, NULL, NULL, NULL) ;
    }
    printf ("ASL time: %e\n", Timer () - tic) ;
    printf ("Total time (read+matrix+ASL): %e\n", Timer () - tic0) ;
    printf ("Time in evaluator: %e\n", PB.val_time) ;
    printf ("Function value from stat: %e\n", Stat.f) ;
    ncol = 0 ;
    nfree = 0 ;
    nhi = 0 ;
    nlo = 0 ;
    atx = 0. ;
    /* count bound and free components of x */
    for (j = 0; j < n; j++)
    {
        /* reverse the transformation on x */
        if ( PB.a [j] < 0 ) x [j] = C - x [j] ;
        if      ( x [j] < 1.e-6 )     nlo++ ;
        else if ( x [j] > C - 1.e-6 ) nhi++ ;
        else                          nfree++ ;
        atx += PB.a [j]*x [j] ;
        /* count number of evaluated columns if problem nonlinear */
        if ( strcmp (argv [2],"linear") ) if ( PB.AP [j] >= 0 ) ncol++ ;
    }
    /* compute objective function value at x */
    scalar_init (g0, 0, n) ;
    if ( !strcmp (argv [2],"linear") ) scalar_init (temp, 0, nfeature) ;
    for (j = 0; j < n; j++)
    {
        if ( x [j] != ZERO )
        {
            xj = .5*x [j] ;
            if ( PB.a [j] < 0. ) xj = -xj ; /* a [j] enters the columns */
            if ( !strcmp (argv [2],"linear") )
            {
                for (k = Ap [j]; k < Ap [j+1]; k++)
                {
                    temp [Ai [k]] += xj*Ax [k] ;
                }
            }
            else
            {
                if ( PB.AP [j] >= 0 ) dense_daxpy (g0, A+(PB.AP [j]*N), xj, n) ;
                else printf ("j: %i xj: %e\n", j, 2*fabs(xj)) ;
            }
        }
    }
    if ( !strcmp (argv [2],"linear") ) /* complete the evaluation of Ax */
    {
        for (j = 0; j < n; j++)
        {
            t = 0 ;
            for (k = Ap [j]; k < Ap [j+1]; k++) t += Ax [k]*temp [Ai [k]] ;
            g0 [j] = t ;
        }
    }
    printf ("nlo: %i nhi: %i nfree: %i\n", nlo, nhi, nfree) ;
    printf ("linear constraint violation: %e\n", atx) ;
    printf ("number of evaluated columns: %i\n", ncol) ;
    /* account for linear term and -1s in quadratic part of cost */
    t = ZERO ;
    for (j = 0; j < n; j++)
    {
        if ( PB.a [j] < 0 ) t += (-1 - g0 [j])*x [j] ;
        else                t += (-1 + g0 [j])*x [j] ;
        
    }
    printf ("FINAL COST: %e\n", t) ;
    free (PB.Ax) ;
    free (PB.Ai) ;
    free (PB.Ap) ;
    free (PB.AP) ;
    free (PB.a) ;
    free (x) ;
    free (g0) ;
    if ( !strcmp (argv [2],"linear") )
    {
        free (PB.temp) ;
    }
    else
    {
        free (PB.A) ;
    }
    i = argc ;
    return 0 ;
    return (1) ;
}

/* sparse dot product x'y */
float sparse_dot
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float t ;
    t = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t += y [j]*x [i] ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    return (t) ;
}

/* binary sparse dot product x'y (entries in x and y are 0 or 1) */
float binary_sparse_dot
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float t ;
    t = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t++ ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    return (t) ;
}

/* ||x_i - x_j||^2 */
float sparse_norm2
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float s, t ;
    s = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t = y [j] - x [j] ;
            s += t*t ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] )
            {
                t = x [i] ;
                s += t*t ;
                i++ ;
            }
            else
            {
                t = y [j] ;
                s += t*t ;
                j++ ;
            }
        }
    }
    if ( i < nzx )
    {
        for (; i < nzx; i++)
        {
            t = x [i] ;
            s += t*t ;
        }
    }
    else
    {
        while ( j < nzy )
        {
            t = y [j] ;
            s += t*t ;
            j++ ;
        }
    }
    return (s) ;
}
/* ||x_i - x_j||^2 with x_i and x_j  binary */
float binary_sparse_norm2
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float s ;
    s = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            i++ ;
            j++ ;
        }
        else
        {
            s++ ;
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    s += (float) ((nzx-i) + (nzy-j)) ;
    return (s) ;
}

void scalar_init
(
    double  *x,  /* numerical values */
    double   s,  /* scalar */
    int      n   /* length of vectors */
)
{
    int j, n5 ;
    double *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

void int_init
(
    int *x,  /* numerical values */
    int  s,  /* scalar */
    int  n   /* length of vectors */
)
{
    int j, n5 ;
    int *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

/* =========================================================================
   === dense_daxpy =========================================================
   =========================================================================
       set x = x + s*y
   ========================================================================= */
void dense_daxpy
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
)
{
    int j, n5 ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] += s*y [j] ;
    for (; j < n; j += 5)
    {
        x [j  ] += s*y [j  ] ;
        x [j+1] += s*y [j+1] ;
        x [j+2] += s*y [j+2] ;
        x [j+3] += s*y [j+3] ;
        x [j+4] += s*y [j+4] ;
    }
}

void get_col
(
    int k   /* build k-th column of A */
)
{
    int j, n, nzx, nzy, *Ap ;
    short degree, *Ai, *Aij ;
    float gamma, t, *A, *Ax, *Axj ;
    long K, N ;
    n = PB.n ;
    N = (long) n ;
    PB.AP [k] = PB.mem ;
    K = N*PB.mem ;
    PB.mem++ ;
    Ap = PB.Ap ;
    Ai = PB.Ai ;
    Ax = PB.Ax ;
    A = PB.A ;
    if ( PB.kernel == 1 ) /* poly */
    {
        gamma = PB.gamma ;
        degree = PB.degree ;
        if ( PB.Ax_is_one )
        {
            nzx = Ap [k+1] - Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = binary_sparse_dot (Aij, nzx, Ai+Ap[j], nzy) ;
                t = gamma*t ;
                t = t*t*t ;
                A [K] = t ;
                K++ ;
            }
        }
        else
        {
            nzx = Ap [k+1] - Ap [k] ;
            Axj = Ax+Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = sparse_dot (Axj, Aij, nzx, Ax+Ap[j], Ai+Ap[j], nzy);
                t = gamma*t ;
                t = t*t*t ;
                A [K] = t ;
                K++ ;
            }
        }
    }
    else if ( PB.kernel == 2 ) /* rbf */
    {
        gamma = PB.gamma ;
        if ( PB.Ax_is_one )
        {
            nzx = Ap [k+1] - Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = binary_sparse_norm2 (Aij, nzx, Ai+Ap[j], nzy) ;
                t = exp (-gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
        else
        {
            nzx = Ap [k+1] - Ap [k] ;
            Axj = Ax+Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = sparse_norm2 (Axj, Aij, nzx, Ax+Ap[j],Ai+Ap[j],nzy);
                t = exp (-gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
    }
    else                       /* sig */
    {
        gamma = PB.gamma ;
        if ( PB.Ax_is_one )
        {
            nzx = Ap [k+1] - Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = binary_sparse_dot (Aij, nzx, Ai+Ap[j], nzy) ;
                t = tanh (gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
        else
        {
            nzx = Ap [k+1] - Ap [k] ;
            Axj = Ax+Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = sparse_dot (Axj, Aij, nzx, Ax+Ap[j], Ai+Ap[j], nzy);
                t = tanh (gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
    }
}

/* SVM nonlinear Ad */
void SVMnonlinearAd
(
    ASLobjective *data
)
{
    int j, k, nz, n, *i, *AP ;
    double tic, *d, *p ;
    float *A ;
    long N ;
    tic = Timer () ;
    nz = data->nz ;
    i = data->i ;
    d = data->d ;
    p = data->p ;
    AP = PB.AP ;
    A = PB.A ;
    n = PB.n ;
    N = n ;
    scalar_init (p, ZERO, n) ;
    for (k = 0; k < nz; k++)
    {
        j = i [k] ;
        if ( AP [j] < 0 ) get_col (j) ;
        dense_daxpy (p, A+(AP [j]*N), d [k], n) ;
    }
    PB.val_time += Timer () - tic ;
}

/* SVM linear Ad */
void SVMlinearAd
(
    ASLobjective *data
)
{
    int j, k, l, m, n, nz, *i, *Ap, *Ap1 ;
    double dk, t, tic, *d, *p, *temp ;
    float *Ax ;
    short *Ai ;
    tic = Timer () ;
    n = PB.n ;
    Ai = PB.Ai ;
    Ap = PB.Ap ;
    Ap1= Ap+1 ;
    Ax = PB.Ax ;
    temp = PB.temp ;
    nz = data->nz ;
    i = data->i ;
    d = data->d ;
    p = data->p ;
    scalar_init (temp, 0., PB.nfeature) ;
    if ( PB.Ax_is_one )
    {
        for (m = 0; m < nz; m++)
        {
            dk = d [m] ;
            j = i [m] ;
            l = Ap1 [j] ;
            for (k = Ap [j]; k < l; k++) temp [Ai [k]] += dk ;
        }
        k = 0 ;
        for (j = 0; j < n; j++)
        {
            t = ZERO ;
            l = Ap1 [j] ;
            for (; k < l; k++) t += temp [Ai [k]] ;
            p [j] = t ;
        }
    }
    else
    {
        for (m = 0; m < nz; m++)
        {
            dk = d [m] ;
            j = i [m] ;
            l = Ap1 [j] ;
            for (k = Ap [j]; k < l; k++) temp [Ai [k]] += dk*Ax [k] ;
        }
        k = 0 ;
        for (j = 0; j < n; j++)
        {
            t = ZERO ;
            l = Ap1 [j] ;
            for (; k < l; k++) t += temp [Ai [k]]*Ax [k] ;
            p [j] = t ;
        }
    }
    PB.val_time += Timer () - tic ;
}

/* read in a problem (in svmlight format) */
void read_problem
(
    const char *filename
)
{
    int elements, i, j, k, max_index, n ;
    float *Ax, Aj ;
    int *Ap ;
    short *a, *Ai, Ax_is_one ;
    FILE *fp = fopen(filename,"r");

    if(fp == NULL)
    {
        fprintf(stderr,"can't open input file %s\n",filename);
        exit(1);
    }

    n = 0 ;
    elements = 0 ;
    while(1)
    {
        int c = fgetc(fp) ;
        switch(c)
        {
            case '\n':
                ++n ; break ; 
            case ':':
                ++elements ; break;
            case EOF:
                goto out;
            default:
            ;
        }
    }
out:
    rewind(fp);

    PB.a = a = (short *) malloc (n*sizeof (short));
    PB.Ax = Ax = (float  *) malloc (elements*sizeof (float)) ;
    PB.Ai = Ai = (short *) malloc (elements*sizeof (short)) ;
    PB.Ap = Ap = (int *) malloc ((n+1)*sizeof (int)) ;
    PB.n = n ;

    Ap [0] = 0 ;
    max_index = 0 ;
    Ax_is_one = TRUE ;
    j = 0 ;
    for(i = 0; i < n; i++)
    {
        fscanf(fp,"%d", &k);
        a [i] = k ;

        while(1)
        {
            int c;
            do {
                c = getc(fp) ;
                if ( c=='\n' ) goto out2 ;
            } while (isspace(c)) ;
            ungetc(c,fp);
            if ( fscanf(fp,"%d:%f", &k, &Aj) < 2 )
            {
              fprintf(stderr,"Wrong input format at line %d\n", i+1) ;
              exit(1) ;
            }
            Ai [j] = k-1 ;
            if ( Aj != 1 ) Ax_is_one = FALSE ;
            Ax [j] = Aj ;
            j++ ;
        }
out2:
        if ( (j >= 1) && k > max_index ) max_index = k ;
        Ap [i+1] = j ;
    }
    PB.nfeature = max_index ;
    PB.Ax_is_one = Ax_is_one ;
    fclose(fp);
}

double Timer ( void )
{
    struct rusage ru ;
    double user, tsys ;

    (void) getrusage (RUSAGE_SELF, &ru) ;

    user =
    ru.ru_utime.tv_sec                  /* user time (seconds) */
    + 1e-6 * ru.ru_utime.tv_usec ;      /* user time (microseconds) */

    tsys =
    ru.ru_stime.tv_sec                  /* tsys time (seconds) */
    + 1e-6 * ru.ru_stime.tv_usec ;      /* tsys time (microseconds) */

    return (user + tsys) ;
}
